import tempfile
from pathlib import Path

import click
from dynaconf import loaders

from auto_proto.core.build import BuildEnv
from auto_proto.core.executor import get_executor
from auto_proto.settings import settings

from ..core import Playbooks
from .exceptions import catch_exception


@click.group()
def show():
    """Show information about the auto-proto project"""
    pass


@show.command()
@click.option(
    "-o",
    "--output-format",
    default="toml",
    help="Output format",
    type=click.Choice(["toml", "json", "yaml"]),
)
def config(output_format):
    """Show the auto-proto project's configuration"""
    with tempfile.TemporaryDirectory() as tmpdir:
        exported_file = tmpdir + "/export"
        data = {key.lower(): val for key, val in settings.as_dict().items()}

        if output_format == "json":
            loaders.json_loader.write(exported_file, data)
        elif output_format == "yaml":
            loaders.yaml_loader.write(exported_file, data)
        else:
            try:
                loaders.toml_loader.write(exported_file, data)
            except ValueError:  # pragma: no cover
                # Some old versions of toml fails because circular references
                print("Dump config to toml failed, try with other format")
                return

        with open(exported_file) as fd:
            print(fd.read())


@show.command(name="manifest")
@catch_exception
def manifest():
    """Show the automotive-image-builder manifest rendered"""
    with tempfile.TemporaryDirectory() as tmpdir:
        build_env = BuildEnv(None, osbuild=True)
        build_env["build"].update({"tmp_dir": tmpdir})

        executor = get_executor(host_id=None)
        with executor() as ctx:
            ctx.run_playbook(Playbooks.SHOW_AIB_MANIFEST, extra_vars=build_env)

        with open(Path(tmpdir, "image.aib.yml")) as fd:
            print(fd.read())
